/*
 * GKRELLM krell for ANTS
 *
 * $Id: antskrell.c,v 1.5 2004/02/16 20:26:45 jakob Exp $
 *
 */

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gkrellm2/gkrellm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "../src/antslib.h"

#define PLUGIN_PLACEMENT 

static void create_plugin(GtkWidget *vbox, gint first_create);
static void update_plugin(void);
static void reset_local_connection(void);
static void handle_local_read(gpointer data, gint source, GdkInputCondition cond);
static void render_status(void);
static void draw_status(void);


static GkrellmPanel *panel;
static GkrellmStyle *style;
static GkrellmTextstyle *ts;
static GkrellmTextstyle *ts_alt;
static GtkWidget *my_vbox;

static GkrellmMonitor plugin_mon = {
  "antsd",
  0,
  create_plugin,
  update_plugin,
  0, /* create_plugin_tab */
  0, /* apply_plugin_config */
  0, /* save_plugin_config */
  0, /* load_plugin_config */
  "antsdkrell",
  0, 0, 0, /* three undefined */
  (MON_NET | MON_INSERT_AFTER),
  0,
  0
};

/*
 * ANTSd communication stuff
 */

static enum {
  STATE_Connecting,
  STATE_Connected
} krell_state = STATE_Connecting;

static gint usock_tag;
static int ants_fd = -1;


/*
 * Jobtype arrays
 */
struct peer_t {
  unsigned peer_id; /* key */
  GkrellmDecal *decal_pt;
  GkrellmDecal **decal_slots;
  unsigned max, preferred;
  unsigned running, reserved;
  struct peer_t *next;
};
struct jobstat_t {
  char jobtype[TYPE_LEN]; /* key */
  GkrellmDecal *decal_jt;
  struct peer_t *peers;
  struct jobstat_t *next;
};
struct jobstat_t *jobs = 0;
static int got_new_data = 0;
static gint render_jobstat(struct jobstat_t*, gint);
static void clear_jobstat(struct jobstat_t *job, int do_dealloc);
static void draw_jobstat(struct jobstat_t *job);

static void insert_jobstat(const char *jobtype,
			   const struct peer_t *peerdata)
{
  struct jobstat_t *ji;
  struct peer_t *pi;
  /* look for jobtype */
  for (ji = jobs; ji && strcmp(ji->jobtype, jobtype); ji = ji->next);
  /* if not found, create */
  if (!ji) {
    struct jobstat_t *newj;
    newj = malloc(sizeof(*newj));
    got_new_data = 1;
    /* init */
    strcpy(newj->jobtype, jobtype);
    newj->decal_jt = 0;
    newj->peers = 0;
    newj->next = jobs;
    jobs = ji = newj;
  }
  /* found... one way or the other */
  /* now look for peer_id */
  for (pi = ji->peers; pi && pi->peer_id != peerdata->peer_id; pi = pi->next);
  /* if not found, create */
  if (!pi) {
    struct peer_t *newp;
    newp = malloc(sizeof(*newp));
    got_new_data = 1;
    /* init */
    newp->decal_pt = 0;
    newp->decal_slots = 0;
    newp->max = -1;
    newp->peer_id = peerdata->peer_id;
    newp->next = ji->peers;
    ji->peers = pi = newp;
  }
  /* found, again one way or the other... */
  /* set values */
  if (pi->max != peerdata->max) {
    /* reallocate decal_slots array array */
    if (pi->decal_slots) {
      unsigned slot;
      for (slot = 0; slot != pi->max; ++slot) {
	gkrellm_remove_decal(panel, pi->decal_slots[slot]);
	gkrellm_destroy_decal(pi->decal_slots[slot]);
      }
      free(pi->decal_slots);
    }
    pi->decal_slots = malloc(sizeof(GkrellmDecal*) * peerdata->max);
    memset(pi->decal_slots, 0, sizeof(GkrellmDecal*) * peerdata->max);
  }
  pi->max = peerdata->max;
  pi->preferred = peerdata->preferred;
  pi->running = peerdata->running;
  pi->reserved = peerdata->reserved;
}


/*
 * Plugin init
 */
GkrellmMonitor *gkrellm_init_plugin(void)
{
  return &plugin_mon;
}

static gint panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		  panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
		  ev->area.width, ev->area.height);
  return FALSE;
}

static void create_plugin(GtkWidget *vbox, gint first_create)
{
  struct jobstat_t *ji;

  my_vbox = vbox;

  /* At first create, allocate the panel struct */
  if (first_create)
    panel = gkrellm_panel_new0();

  /* Clear any decals we may have in local data */
  for (ji = jobs; ji; ji = ji->next)
    clear_jobstat(ji, 0);

  /* Use the default style */
  style = gkrellm_meter_style(DEFAULT_STYLE_ID);

  /* We use the 'alt' text style because it is presumably smaller than
     the default text style */
  ts = gkrellm_meter_textstyle(DEFAULT_STYLE_ID);
  ts_alt = gkrellm_meter_alt_textstyle(DEFAULT_STYLE_ID);

  /* Render... */
  render_status();

  /* Calculate size of panel */
  gkrellm_panel_configure(panel, 0, style);

  /* Create the panel */
  gkrellm_panel_create(vbox, &plugin_mon, panel);

  /* If first create, init. antsd connection */
  if (first_create) {
    ants_fd = ants_local_connect();
    /* Also let the local antsd know that we want status updates */
    ants_statclient(ants_fd);
    /* Connect read events on the status fd to handler func. */
    usock_tag = gdk_input_add(ants_fd, GDK_INPUT_READ,
			      handle_local_read, 0);
    krell_state = STATE_Connected;
  }

  /* If first create, connect expose event */
  if (first_create) {
    gtk_signal_connect(GTK_OBJECT(panel->drawing_area), "expose_event",
		       (GtkSignalFunc)panel_expose_event, 0);

  }
}


static void update_plugin(void)
{
  switch (krell_state) {
  case STATE_Connecting: {
    /* In connecting state, print that... */
/*     gkrellm_draw_decal_text(panel, conn_text, "No ANTS", -1); */
    break;
  }
  case STATE_Connected: {
    /* In normal state, render the entire status stuff */
    if (got_new_data) {
      /* re-render as we got new data */
      render_status();
      got_new_data = 0;
    }
    draw_status();
    break;
  }
  default:
    fprintf(stderr, "Huh!\n");
  }
  gkrellm_draw_panel_layers(panel);
}

static void handle_local_read(gpointer data, gint source, GdkInputCondition cond)
{
  static char * buf = 0;
  int rc;

  struct peer_t peer;
  char jobtype[TYPE_LEN];

  /*
   * Make sure we're supposed to read...
   */
  if(cond != GDK_INPUT_READ) {
    reset_local_connection();
    return;
  }
  /*
   * Check buffer alloc.
   */
  if(!buf) {
    buf = (char*)malloc(LOG_LEN);
    if(!buf) return;
  }
  /*
   * Read back status
   */  
  rc = recv(ants_fd, buf, LOG_LEN, 0);
  if(rc <= 0) {
    reset_local_connection();
    return;
  }

  rc = sscanf(buf, "%u %15s %u %u %u %u",
	      &peer.peer_id, jobtype, 
	      &peer.max, &peer.preferred,
	      &peer.running, &peer.reserved);
  if(rc != 6) {
    g_print(_("Yuck! sscanf() failed\n"));
    return;
  }
  /* Update stats */
  insert_jobstat(jobtype, &peer);
  /* That's it - update now */
  update_plugin();
}

static void reset_local_connection(void)
{
  /* remove gdk tag */
  gdk_input_remove(usock_tag);
  /* close descriptor */
  close(ants_fd);
  /* get new descriptor */
  ants_fd = ants_local_connect();
  /* notify local antsd */
  ants_statclient(ants_fd);
  /* add input again */
  usock_tag = gdk_input_add(ants_fd, GDK_INPUT_READ,
			    handle_local_read, 0);
}

static void draw_status(void)
{
  /* For each jobtype, draw it */
  struct jobstat_t *ji;
  for (ji = jobs; ji; ji = ji->next) {
    draw_jobstat(ji);
  }
}

static void render_status(void)
{
  struct jobstat_t *ji;
  gint ypos = 4;
  /* For each jobtype, render it */
  for (ji = jobs; ji; ji = ji->next) {
    /* Clear all locally held decals - if any */
    clear_jobstat(ji, 1);
    /* Now render */
    ypos = render_jobstat(ji, ypos);
  }
  /* notify gkrellm of layout changes */
  gkrellm_panel_configure(panel, 0, style);
  gkrellm_panel_create(my_vbox, &plugin_mon, panel);
}

static gint render_peerstat(struct peer_t *peer, gint ypos)
{
  unsigned slot;
  GdkPixmap *warn_pix;
  GdkBitmap *warn_map;
  gint xpos, ypos_tmp = ypos;
  /* Print peer name, slightly indented */
  peer->decal_pt = gkrellm_create_decal_text(panel, "Ay", ts_alt, style, 6, ypos, 8);
  xpos = peer->decal_pt->w + 2;
  /* Init lamps */
  warn_pix = gkrellm_decal_misc_pixmap();
  warn_map = gkrellm_decal_misc_mask();
  /* For each jobslot, render a lamp */
  for (slot = 0; slot != peer->max; ++slot) {
    /* if x is too far to the right, wrap */
    if (xpos + 14 > gkrellm_chart_width()) {
      xpos = peer->decal_pt->w + 2;
      ypos_tmp += 6;
    }
    /* now render */
    peer->decal_slots[slot]
      = gkrellm_create_decal_pixmap(panel, warn_pix, warn_map,
				    N_MISC_DECALS, 0,
				    xpos,
				    ypos_tmp);
    /* update x position */
    xpos += 6;
  }
  ypos += peer->decal_pt->h + 2;
  return ypos;
}

static void clear_peerstat(struct peer_t *peer, int do_dealloc)
{
  unsigned slot;
  /* clear peerstat */
  if (peer->decal_pt) {
    if (do_dealloc) {
      gkrellm_remove_decal(panel, peer->decal_pt);
      gkrellm_destroy_decal(peer->decal_pt);
    }
    peer->decal_pt = 0;
  }
  /* For each jobslot, clear the lamp decal */
  for (slot = 0; slot != peer->max; ++slot) {
    if (peer->decal_slots[slot]) {
      if (do_dealloc) {
	gkrellm_remove_decal(panel, peer->decal_slots[slot]);
	gkrellm_destroy_decal(peer->decal_slots[slot]);
      }
      peer->decal_slots[slot] = 0;
    }
  }
}

static gint render_jobstat(struct jobstat_t *job, gint ypos)
{
  struct peer_t *pi;
  /* Print pretty jobtype text */
  job->decal_jt = gkrellm_create_decal_text(panel, "Ay", ts, style, 4, ypos, -1);
  ypos += job->decal_jt->h + 2;

  /* Print each peer */
  for (pi = job->peers; pi; pi = pi->next)
    ypos = render_peerstat(pi, ypos);

  return ypos;
}

static void clear_jobstat(struct jobstat_t *job, int do_dealloc)
{
  struct peer_t *pi;
  /* Clear our decal if any */
  if (job->decal_jt) {
    if (do_dealloc) {
      gkrellm_remove_decal(panel, job->decal_jt);
      gkrellm_destroy_decal(job->decal_jt);
    }
    job->decal_jt = 0;
  }

  /* Clear on each peer */
  for (pi = job->peers; pi; pi = pi->next)
    clear_peerstat(pi, do_dealloc);
}

static void draw_peerstat(struct peer_t *peer)
{
  unsigned slot;
  /* print the peer decals */
  gkrellm_draw_decal_text(panel, peer->decal_pt, "p:", -1);
  for (slot = 0; slot != peer->max; ++slot)
    gkrellm_draw_decal_pixmap(panel, peer->decal_slots[slot],
			      peer->running > slot ? 5 : 4);
}

static void draw_jobstat(struct jobstat_t *job)
{
  struct peer_t *pi;
  /* Print pretty jobtype text */
  gkrellm_draw_decal_text(panel, job->decal_jt, job->jobtype, -1);
  /* Print each peer */
  for (pi = job->peers; pi; pi = pi->next)
    draw_peerstat(pi);
}
